//cs599 Homework 2 Problem 2
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>

#define N 1000

void performConvolution(int **, int **, int [][3]);
void printPartialImage(int **);

int main(int argc, char ** argv)
{
  int ** myImage;
  int ** horizontalEdges;
  int ** verticalEdges;

  int vOp[3][3] = { {-1, -2, -1},
                    {0, 0, 0},
                    {1, 2, 1} };
  int hOp[3][3] = { {-1, 0, 1},
                    {-2, 0, 2},
                    {-1, 0, 1} };
  int i, j, k;

  myImage = (int **) malloc(sizeof(int *)*N);
  horizontalEdges = (int **) malloc(sizeof(int *)*N);
  verticalEdges = (int **) malloc(sizeof(int *)*N);

  //Generate 1000 by 1000 2D arrays
  for(i = 0; i < N; i++)
  {
    myImage[i] = (int *) malloc(sizeof(int)*N);
    horizontalEdges[i] = (int *) malloc(sizeof(int)*N);
    verticalEdges[i] = (int *) malloc(sizeof(int)*N);
  }

  //fill the array
  for(i = 0; i < N; i++)
    for(j = 0; j < N; j++)
      myImage[i][j] = i;

  //Perform horizonal edge detection
  performConvolution(myImage, horizontalEdges, hOp);
  
  //Perform vertical edge detection
  performConvolution(myImage, verticalEdges, vOp);

  //Print part of each image:
  printf("Original\n");
  printPartialImage(myImage);
  printf("Vertical edges\n");
  printPartialImage(verticalEdges);
  printf("Horizontal edges\n");
  printPartialImage(horizontalEdges);

  //Free data
  for(i = 0; i < N; i++)
  {
    free((void *) myImage[i]);
    free((void *) verticalEdges[i]);
    free((void *) horizontalEdges[i]);
  }

  free((void *) myImage);
  free((void *) verticalEdges);
  free((void *) horizontalEdges);
}

void printPartialImage(int ** arr)
{
  int i, j;
  //Print the first two lines and last two lines
  for(i = 0; i < 2; i++)
  {
    for(j = 0; j < 40; j++)
      printf("%d ", arr[i][j]);
    for(j = N-40; j < N; j++)
      printf("%d ", arr[i][j]);
    printf("\n");
  }

  for(i = N-2; i < N; i++)
  {
    for(j = 0; j < 40; j++)
      printf("%d ", arr[i][j]);
    for(j = N-40; j < N; j++)
      printf("%d ", arr[i][j]);
    printf("\n");
  }
}

void performConvolution(int ** inputImage, int ** outputImage, int operator[][3])
{
  int i, j;

  #pragma omp parallel for private(i,j) shared(inputImage, outputImage, operator)
  for(i = 0; i < N; i++)
  {
    for(j = 0; j < N; j++)
    {
      int i_1 = i-1 < 0 ? N-1 : i-1;
      int j_1 = j-1 < 0 ? N-1 : j-1;
      int ip1 = i+1 >= N ? 0 : i+1;
      int jp1 = j+1 >= N ? 0 : j+1;
      outputImage[i][j] = 
       inputImage[i_1][j_1]*operator[0][0] +
       inputImage[i_1][j] * operator[0][1] +
       inputImage[i_1][jp1]*operator[0][2] +
       inputImage[i][j_1] * operator[1][0] +
       inputImage[i][j]  *  operator[1][1] +
       inputImage[i][jp1] * operator[1][2] +
       inputImage[ip1][j_1]*operator[2][0] +
       inputImage[ip1][j] * operator[2][1] +
       inputImage[ip1][jp1]*operator[2][2];
    }
  }
}
