//HW2 Problem 5 Solution
//This solution is based upon code from the following site:
//  http://www.oscer.ou.edu/Workshops/StorageHierarchy/sipe_storage_20150127.pdf

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

void matrixMultiplyByTiling(double **, double **, double **, int, int, int, int, int, int);
void tileMultiply(double **, double **, double **, int, int, int, int, int, int);

#define N 2000
#define NUM_THREADS 8

int main(int argc, char ** argv)
{
  double ** A;
  double ** B;
  double ** C;

  int i, j;

  A = (double **) malloc(sizeof(double *)*N);
  B = (double **) malloc(sizeof(double *)*N);
  C = (double **) malloc(sizeof(double *)*N);

  for(i = 0; i < N; i++)
  {
    A[i] = (double *) malloc(sizeof(double)*N);
    B[i] = (double *) malloc(sizeof(double)*N);
    C[i] = (double *) malloc(sizeof(double)*N);
  }

  for(i = 0; i < N; i++)
    for(j = 0; j < N; j++)
      A[i][j] = B[i][j] = i;

  time_t start = time(NULL);
  clock_t start_c = clock();
  matrixMultiplyByTiling(C, A, B, N, N, N, 40, 40, 40);
  time_t end = time(NULL);
  clock_t end_c = clock();

  printf("Wall time is %lf\n", (double)difftime(end,start));
  printf("Approximate CPU time is %lf\n", ((double)end_c-start_c)/(CLOCKS_PER_SEC*NUM_THREADS));


  for(i = 0; i < 2; i++)
  {
    for(j = 0; j < N; j++)
      printf("%lf ", C[i][j]);
    printf("\n");
  }

  for(i = N-2; i < N; i++)
  {
    for(j = 0; j < N; j++)
      printf("%lf ", C[i][j]);
    printf("\n");
  }

  for(i = 0; i < N; i++)
  {
    free((void *) A[i]);
    free((void *) B[i]);
    free((void *) C[i]);
  }

  free((void *) A);
  free((void *) B);
  free((void *) C);

  return 0;
}

void matrixMultiplyByTiling(double ** dst, double ** src1, double ** src2, 
  int nr, int nc, int nq,
  int rtilesize, int ctilesize, int qtilesize)
{
  int rstart, rend, cstart, cend, qstart, qend;
  for (rstart = 0; rstart < nr; rstart += rtilesize) {
    rend = rstart + rtilesize - 1;
    if (rend >= nr) rend = nr - 1;
    for (cstart = 0; cstart < nc; cstart += ctilesize) {
      cend = cstart + ctilesize - 1;
      if (cend >= nc) cend = nc - 1;
      for (qstart = 0; qstart < nq; qstart += qtilesize) {
        qend = qstart + qtilesize - 1;
        if (qend >= nq) qend = nq - 1;
        tileMultiply(dst, src1, src2,
        rstart, rend, cstart, cend, qstart, qend);
      } /* for qstart */
    } /* for cstart */
  } /* for rstart */
}

void tileMultiply(double ** dst, double ** src1, double ** src2, 
  int rstart, int rend, int cstart, int cend, int qstart, int qend)
{
  int r, c, q;
  #pragma omp parallel for private(r, c, q) shared(dst, src1, src2, rstart, rend, qstart, qend, cstart, cend) num_threads(NUM_THREADS)
  for (r = rstart; r <= rend; r++) {
    for (c = cstart; c <= cend; c++) {      
      if (qstart == 0) dst[r][c] = 0.0;
      double dstrc = dst[r][c];
      for (q = qstart; q <= qend; q++) {
        dstrc += src1[r][q] * src2[q][c];
      } /* for q */
      dst[r][c] = dstrc;
    } /* for c */
  } /* for r */
}
